from pbpstats.resources.enhanced_pbp import StartOfPeriod


class NbaPossessionLoader(object):
    """
    Class for shared methods between :obj:`~pbpstats.data_loader.data_nba.possessions_loader.DataNbaPossessionLoader`
    and :obj:`~pbpstats.data_loader.stats_nba.possessions_loader.StatsNbaPossessionLoader`

    Both :obj:`~pbpstats.data_loader.data_nba.possessions_loader.DataNbaPossessionLoader`
    and :obj:`~pbpstats.data_loader.stats_nba.possessions_loader.StatsNbaPossessionLoader` should inherit from this class

    This class should not be instantiated directly
    """

    def _split_events_by_possession(self):
        """
        splits events by possession
        :returns: list of lists with events for each possession
        """
        events = []
        possession_events = []
        for event in self.events:
            possession_events.append(event)
            if event.is_possession_ending_event:
                events.append(possession_events)
                possession_events = []
        return events

    def _add_extra_attrs_to_all_possessions(self):
        """
        adds possession number and next and previous possession
        """
        number = 1
        for i, possession in enumerate(self.items):
            if i == 0 and i == len(self.items) - 1:
                possession.previous_possession = None
                possession.next_possession = None
            elif isinstance(possession.events[0], StartOfPeriod) or i == 0:
                possession.previous_possession = None
                possession.next_possession = self.items[i + 1]
                number = 1
            elif (
                i == len(self.items) - 1
                or possession.period != self.items[i + 1].period
            ):
                possession.previous_possession = self.items[i - 1]
                possession.next_possession = None
            else:
                possession.previous_possession = self.items[i - 1]
                possession.next_possession = self.items[i + 1]
            possession.number = number
            number += 1



"""
The ``EnhancedPbp`` class has some basic properties for handling enhanced pbp data
"""
from pbpstats.resources.base import Base
from pbpstats.resources.enhanced_pbp import FieldGoal, FreeThrow, Rebound, Turnover


class EnhancedPbp(Base):
    """
    Class for enhanced play-by-play events

    :param list events: list of
        :obj:`~pbpstats.resources.enhanced_pbp.enhanced_pbp_item.EnhancedPbpItem` items,
        typically from a enhanced pbp data loader
    """

    def __init__(self, items):
        self.items = items

    @property
    def data(self):
        """
        returns list of pbp event dicts
        """
        return [item.data for item in self.items]

    @property
    def fgas(self):
        """
        returns list of :obj:`~pbpstats.resources.enhanced_pbp.field_goal.FieldGoal` events
        """
        return [item for item in self.items if isinstance(item, FieldGoal)]

    @property
    def fgms(self):
        """
        returns list of :obj:`~pbpstats.resources.enhanced_pbp.field_goal.FieldGoal` events with all made FGs
        """
        return [
            item for item in self.items if isinstance(item, FieldGoal) and item.is_made
        ]

    @property
    def ftas(self):
        """
        returns list of :obj:`~pbpstats.resources.enhanced_pbp.free_throw.FreeThrow` events
        """
        return [item for item in self.items if isinstance(item, FreeThrow)]

    @property
    def rebounds(self):
        """
        returns list of :obj:`~pbpstats.resources.enhanced_pbp.field_goal.Rebound` events
        """
        return [
            item
            for item in self.items
            if isinstance(item, Rebound) and item.is_real_rebound
        ]

    @property
    def turnovers(self):
        """
        returns list of :obj:`~pbpstats.resources.enhanced_pbp.field_goal.Turnover` events
        """
        return [
            item
            for item in self.items
            if isinstance(item, Turnover) and not item.is_no_turnover
        ]



from pbpstats import DEFENSIVE_GOALTENDING_STRING


class Violation(object):
    """
    Class for violation events
    """

    event_type = 7

    @property
    def is_delay_of_game(self):
        return self.event_action_type == 1

    @property
    def is_goaltend_violation(self):
        return self.event_action_type == 2

    @property
    def is_lane_violation(self):
        return self.event_action_type == 3

    @property
    def is_jumpball_violation(self):
        return self.event_action_type == 4

    @property
    def is_kicked_ball_violation(self):
        return self.event_action_type == 5

    @property
    def is_double_lane_violation(self):
        return self.event_action_type == 6

    @property
    def event_stats(self):
        """
        returns list of dicts with all stats for event
        """
        stats = []
        if self.is_goaltend_violation:
            stats.append(
                {
                    "player_id": self.player1_id,
                    "team_id": self.team_id,
                    "stat_key": DEFENSIVE_GOALTENDING_STRING,
                    "stat_value": 1,
                }
            )
            team_ids = list(self.current_players.keys())
            lineups_ids = self.lineup_ids
            for stat in stats:
                opponent_team_id = (
                    team_ids[0] if stat["team_id"] == team_ids[1] else team_ids[1]
                )
                stat["lineup_id"] = lineups_ids[stat["team_id"]]
                stat["opponent_team_id"] = opponent_team_id
                stat["opponent_lineup_id"] = lineups_ids[opponent_team_id]
        return self.base_stats + stats



class Replay(object):
    """
    Class for replay events
    """

    event_type = 18

    @property
    def support_ruling(self):
        return self.event_action_type == 4

    @property
    def overturn_ruling(self):
        return self.event_action_type == 5

    @property
    def ruling_stands(self):
        return self.event_action_type == 6

    @property
    def event_stats(self):
        """
        returns list of dicts with all stats for event
        """
        return self.base_stats



KEY_ATTR_MAPPER = {
    "GAME_ID": "game_id",
    "GAME_DATE_EST": "date",
    "HOME_TEAM_ID": "home_team_id",
    "VISITOR_TEAM_ID": "visitor_team_id",
    "GAME_STATUS_TEXT": "status",
}


class StatsNbaGameItem(object):
    """
    Class for game data from stats.nba.com

    :param dict item: dict with game data
    """

    def __init__(self, item):
        for key, value in KEY_ATTR_MAPPER.items():
            if item.get(key) is not None:
                setattr(self, value, item.get(key))

    @property
    def data(self):
        """
        returns game dict
        """
        return self.__dict__

    @property
    def is_final(self):
        """
        returns True if game is final, False otherwise
        """
        return self.status == "Final"



"""
``StatsEnhancedPbpItem`` is the base class for all stats.nba.com enhanced pbp event types
"""
from collections import defaultdict

import requests

from pbpstats import HEADERS, REQUEST_TIMEOUT
from pbpstats.resources.enhanced_pbp.enhanced_pbp_item import EnhancedPbpItem
from pbpstats.resources.enhanced_pbp import (
    FieldGoal,
    Foul,
    FreeThrow,
    JumpBall,
    Rebound,
    StartOfPeriod,
    Turnover,
    Violation,
)


KEY_ATTR_MAPPER = {
    "GAME_ID": "game_id",
    "EVENTNUM": "event_num",
    "PCTIMESTRING": "clock",
    "PERIOD": "period",
    "EVENTMSGACTIONTYPE": "event_action_type",
    "EVENTMSGTYPE": "event_type",
    "PLAYER1_ID": "player1_id",
    "PLAYER1_TEAM_ID": "team_id",
    "PLAYER2_ID": "player2_id",
    "PLAYER3_ID": "player3_id",
    "VIDEO_AVAILABLE_FLAG": "video_available",
}


class StatsEnhancedPbpItem(EnhancedPbpItem):
    """
    Base class for enhanced pbp events from stats.nba.com

    :param dict event: dict with event data
    :param int order: sequential order in which event occurs
    """

    def __init__(self, event, order):
        for key, value in KEY_ATTR_MAPPER.items():
            if event.get(key) is not None:
                setattr(self, value, event.get(key))

        if (
            event.get("HOMEDESCRIPTION") is not None
            and event.get("VISITORDESCRIPTION") is not None
        ):
            self.description = (
                f"{event.get('HOMEDESCRIPTION')}: {event.get('VISITORDESCRIPTION')}"
            )
        elif event.get("HOMEDESCRIPTION") is not None:
            self.description = f"{event.get('HOMEDESCRIPTION')}"
        elif event.get("VISITORDESCRIPTION") is not None:
            self.description = f"{event.get('VISITORDESCRIPTION')}"
        elif event.get("NEUTRALDESCRIPTION") is not None:
            self.description = f"{event.get('NEUTRALDESCRIPTION')}"
        else:
            self.description = ""

        if event.get("PLAYER1_TEAM_ID") is None and event.get("PLAYER1_ID") is not None:
            # need to set team id in these cases where player id is team id
            self.team_id = event.get("PLAYER1_ID", 0)
            self.player1_id = 0

        # fix team/player ids on some event types so they are consistent with DataPbpItem
        if self.event_type == 10:
            # jump ball PLAYER3_TEAM_ID is player who ball gets tipped to
            self.player2_id = event["PLAYER3_ID"]
            self.player3_id = event["PLAYER2_ID"]
            if event["PLAYER3_TEAM_ID"] is not None:
                self.team_id = event["PLAYER3_TEAM_ID"]
            else:
                # when jump ball is tipped out of bounds, winning team is PLAYER3_ID
                self.team_id = event["PLAYER3_ID"]
                if hasattr(self, "player2_id"):
                    delattr(self, "player2_id")
        elif self.event_type in [5, 6]:
            # steals need to change PLAYER2_ID to player3_id - this is player who turned ball over
            # fouls need to change PLAYER2_ID to player3_id - this is player who drew foul
            if hasattr(self, "player2_id"):
                delattr(self, "player2_id")
            if event.get("PLAYER2_ID") is not None:
                self.player3_id = event["PLAYER2_ID"]

        if hasattr(self, "player2_id") and self.player2_id == 0:
            delattr(self, "player2_id")
        if hasattr(self, "player3_id") and self.player3_id == 0:
            delattr(self, "player3_id")
        self.order = order
        self.player_game_fouls = defaultdict(int)
        self.possession_changing_override = False
        self.non_possession_changing_override = False
        self.score = defaultdict(int)

    @property
    def data(self):
        """
        returns event as a dict
        """
        return self.__dict__

    @property
    def video_url(self):
        """
        returns url for mp4 video of play, if available
        """
        if self.video_available == 1:
            parameters = {"GameEventID": self.event_num, "GameID": self.game_id}
            base_url = "https://stats.nba.com/stats/videoeventsasset"
            response = requests.get(
                base_url, parameters, headers=HEADERS, timeout=REQUEST_TIMEOUT
            )
            if response.status_code == 200:
                response_json = response.json()
                video_urls = response_json["resultSets"]["Meta"]["videoUrls"]
                if len(video_urls) == 1:
                    return video_urls[0]["murl"]
            else:
                response.raise_for_status()
        return None

    def get_offense_team_id(self):
        """
        returns team id for team on offense for event
        """
        if isinstance(self, Foul) and (self.is_charge or self.is_offensive_foul):
            # offensive foul returns team id
            # this isn't separate method in Foul class because some fouls can be committed
            # on offense or defense (loose ball, flagrant, technical)
            return self.team_id
        event_to_check = self.previous_event
        team_ids = list(self.current_players.keys())
        while event_to_check is not None and not (
            isinstance(event_to_check, (FieldGoal, JumpBall))
            or (
                isinstance(event_to_check, Turnover)
                and not event_to_check.is_no_turnover
            )
            or (isinstance(event_to_check, Rebound) and event_to_check.is_real_rebound)
            or (
                isinstance(event_to_check, FreeThrow)
                and not event_to_check.is_technical_ft
            )
        ):
            event_to_check = event_to_check.previous_event
        if event_to_check is None and self.next_event is not None:
            # should only get here on first possession of period when first event is non-offensive foul,
            # FieldGoal, FreeThrow, Rebound, Turnover, JumpBall
            return self.next_event.get_offense_team_id()
        if isinstance(event_to_check, Turnover) and not event_to_check.is_no_turnover:
            return (
                team_ids[0]
                if team_ids[1] == event_to_check.get_offense_team_id()
                else team_ids[1]
            )
        if isinstance(event_to_check, Rebound) and event_to_check.is_real_rebound:
            if not event_to_check.oreb:
                return (
                    team_ids[0]
                    if team_ids[1] == event_to_check.get_offense_team_id()
                    else team_ids[1]
                )
            return event_to_check.get_offense_team_id()
        if isinstance(event_to_check, (FieldGoal, FreeThrow)):
            if event_to_check.is_possession_ending_event:
                return (
                    team_ids[0]
                    if team_ids[1] == event_to_check.get_offense_team_id()
                    else team_ids[1]
                )
            return event_to_check.get_offense_team_id()
        if isinstance(event_to_check, JumpBall):
            if event_to_check.count_as_possession:
                team_ids = list(self.current_players.keys())
                return (
                    team_ids[0]
                    if team_ids[1] == event_to_check.get_offense_team_id()
                    else team_ids[1]
                )
            return event_to_check.get_offense_team_id()

    @property
    def is_possession_ending_event(self):
        """
        returns True if event ends a possession, False otherwise
        """
        if self.next_event is None:
            return True

        if self.possession_changing_override:
            return True

        if self.non_possession_changing_override:
            return False

        if isinstance(self, Rebound) and self.is_real_rebound and not self.oreb:
            return True

        if isinstance(self, Turnover) and not self.is_no_turnover:
            return True

        if isinstance(self, FieldGoal) and self.is_made:
            # no possession change on flagrant foul
            next_event_is_flagrant_drawn = (
                isinstance(self.next_event, Foul)
                and self.next_event.is_flagrant
                and self.team_id != self.next_event.team_id
                and self.clock == self.next_event.clock
            )
            if (
                not self.is_make_that_does_not_end_possession
                and not next_event_is_flagrant_drawn
            ):
                return True

        if isinstance(self, FreeThrow) and self.is_made and self.is_end_ft:
            next_event_is_foul_drawn_at_ft_time = (
                isinstance(self.next_event, Foul)
                and self.clock == self.next_event.clock
                and self.team_id != self.next_event.team_id
                and (
                    self.next_event.is_loose_ball_foul
                    or self.next_event.is_personal_foul
                    or self.next_event.is_away_from_play_foul
                    or self.next_event.is_flagrant
                )
            )
            if (
                not self.is_away_from_play_ft
                and not self.is_inbound_foul_ft
                and not next_event_is_foul_drawn_at_ft_time
            ):
                return True

        if not isinstance(self.previous_event, StartOfPeriod) and isinstance(
            self, JumpBall
        ):
            return self._is_jump_ball_possession_ending_event()

        return False

    def _is_jump_ball_possession_ending_event(self):
        """
        need to check for rare case where possession changes on jump ball but there is no turnover/rebound
        """
        if (
            isinstance(self.next_event, Turnover)
            and not self.next_event.is_no_turnover
            and self.next_event.clock == self.clock
        ):
            # if next event is steal at same time of pbp don't need to change possession since steal takes care of it
            return False
        elif (
            isinstance(self.previous_event, Turnover)
            and not self.previous_event.is_no_turnover
            and self.previous_event.clock == self.clock
        ):
            # if previous event is steal at same time of pbp don't need to change possession since steal takes care of it
            return False
        elif (
            isinstance(self.next_event, Violation)
            and self.next_event.is_jumpball_violation
        ):
            # jump ball violation - turnover will be possession changing event
            return False
        elif isinstance(self.next_event, Foul) and self.next_event.clock == self.clock:
            next_event = self.next_event.next_event
            if (
                isinstance(next_event, Turnover)
                and not next_event.is_no_turnover
                and next_event.clock == self.clock
            ):
                # foul turnover on jump ball - turnover will trigger change of possession
                return False

        jump_ball_winning_team_id = self.team_id

        prev_event = self.previous_event
        while prev_event is not None and not prev_event.is_possession_ending_event:
            prev_event = prev_event.previous_event

        if prev_event is None:
            return False

        if isinstance(prev_event, Rebound):
            jump_ball_winning_team_started_possession_with_ball = (
                jump_ball_winning_team_id == prev_event.team_id
            )
        else:
            jump_ball_winning_team_started_possession_with_ball = (
                jump_ball_winning_team_id != prev_event.team_id
            )

        next_event_rebound = (
            isinstance(self.next_event, Rebound) and self.next_event.is_real_rebound
        )

        if not jump_ball_winning_team_started_possession_with_ball and not (
            next_event_rebound or isinstance(self.next_event, JumpBall)
        ):
            # ignore jump ball if next event is a rebound or jump ball since that will trigger possession change
            return True
        return False



from pbpstats.resources.enhanced_pbp.stats_nba.enhanced_pbp_item import (
    StatsEnhancedPbpItem,
)
from pbpstats.resources.enhanced_pbp import Turnover


class StatsTurnover(Turnover, StatsEnhancedPbpItem):
    """
    Class for Turnover events
    """

    def __init__(self, *args):
        super().__init__(*args)

    def get_offense_team_id(self):
        """
        returns team id for team on offense for event
        """
        if self.is_no_turnover:
            return self.previous_event.get_offense_team_id()
        return self.team_id



import json


class IntDecoder(json.JSONDecoder):
    def decode(self, s):
        result = super().decode(s)
        return self._decode(result)

    def _decode(self, o):
        if isinstance(o, str):
            try:
                return int(o)
            except ValueError:
                return o
        elif isinstance(o, dict):
            return {
                convert_to_int_but_keep_game_id_string(k): self._decode(v)
                for k, v in o.items()
            }
        elif isinstance(o, list):
            return [self._decode(v) for v in o]
        else:
            return o


def convert_to_int_but_keep_game_id_string(value):
    try:
        if str(int(value)) == value:
            return int(value)
        else:
            return value
    except ValueError:
        return value



class DataNbaPbpItem(object):
    """
    Class for pbp events from data.nba.com

    :param dict item: dict with event data
    :param int period: period in which event occurs
    """

    def __init__(self, item, period):
        self.period = period
        for key, value in item.items():
            setattr(self, key, value)

    @property
    def data(self):
        """
        returns pbp event dict
        """
        return self.__dict__



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="pbpstats",
    version="1.1.2",
    author="Darryl Blackport",
    author_email="darryl.blackport@gmail.com",
    description="A package to scrape and parse NBA, WNBA and G-League play-by-play data",
    license="MIT License",
    keywords=["basketball", "NBA", "WNBA", "G-League", "play-by-play", "pbp"],
    url="https://github.com/dblackrun/pbpstats",
    packages=find_packages(),
    install_requires=["requests"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



from pbpstats.resources.enhanced_pbp.stats_nba.enhanced_pbp_item import (
    StatsEnhancedPbpItem,
)
from pbpstats.resources.enhanced_pbp import JumpBall, Turnover, Foul


class StatsJumpBall(JumpBall, StatsEnhancedPbpItem):
    """
    Class for jump ball events
    """

    def __init__(self, *args):
        super().__init__(*args)

    def get_offense_team_id(self):
        """
        returns team id for team on offense for event
        """
        if self.next_event.clock == self.clock and isinstance(
            self.next_event, Turnover
        ):
            return self.next_event.team_id
        if isinstance(self.next_event, Foul) and self.next_event.clock == self.clock:
            next_event = self.next_event.next_event
            if (
                isinstance(next_event, Turnover)
                and not next_event.is_no_turnover
                and next_event.clock == self.clock
            ):
                return next_event.team_id
        if self.count_as_possession:
            team_ids = list(self.current_players.keys())
            return team_ids[0] if team_ids[1] == self.team_id else team_ids[1]
        return self.team_id



import abc
from functools import wraps


def check_file_directory(method):
    """
    decorator to be used on method to load data from file to
    check that file directory is not None

    :raises: ValueError: If file_directory is None
    """

    @wraps(method)
    def decorated_method(self, *method_args, **method_kwargs):
        if self.file_directory is None:
            raise ValueError(f"file_directory cannot be None when data source is file")
        return method(self, *method_args, **method_kwargs)

    return decorated_method


class AbsDataLoader(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def _load_data(self):
        pass

    @abc.abstractproperty
    def data(self):
        pass



from pbpstats.resources.enhanced_pbp.ejection import Ejection
from pbpstats.resources.enhanced_pbp.end_of_period import EndOfPeriod
from pbpstats.resources.enhanced_pbp.foul import Foul
from pbpstats.resources.enhanced_pbp.free_throw import FreeThrow
from pbpstats.resources.enhanced_pbp.substitution import Substitution
from pbpstats.resources.enhanced_pbp.turnover import Turnover
from pbpstats.resources.enhanced_pbp.violation import Violation
from pbpstats.resources.enhanced_pbp.field_goal import FieldGoal
from pbpstats.resources.enhanced_pbp.jump_ball import JumpBall
from pbpstats.resources.enhanced_pbp.replay import Replay
from pbpstats.resources.enhanced_pbp.timeout import Timeout
from pbpstats.resources.enhanced_pbp.rebound import Rebound
from pbpstats.resources.enhanced_pbp.start_of_period import (
    StartOfPeriod,
    InvalidNumberOfStartersException,
)

__all__ = [
    "Ejection",
    "EndOfPeriod",
    "FieldGoal",
    "Foul",
    "FreeThrow",
    "JumpBall",
    "Rebound",
    "Replay",
    "StartOfPeriod",
    "InvalidNumberOfStartersException",
    "Substitution",
    "Timeout",
    "Turnover",
    "Violation",
]
